@@ -16,6 +16,7 @@ from xapian import Query, QueryParser, UnimplementedError, DocNotFoundError
 
 from . import snowball
 from .exc import ValidationError, PrefixError
+from .tools import LRUDict
 
 
 RETRY_LIMIT = 5
@@ -56,6 +57,11 @@ def _prefix(name):
     return (u'X%s:' % name.upper()).encode('utf-8')
 
 
+def to_term(value, prefix=None):
+    value = value.lower()
+    return _prefix(prefix) + value if prefix else value
+
+
 class Record(object):
     """Nice attribute-accessable record for a search result."""
 
@@ -191,6 +197,8 @@ class Database(object):
     value_sort_prefix = "_XODB_VALUESORT_"
     value_count_name = "_XODB_COUNT_"
     backend = None
+    _metadata_keyset = None
+    query_cache_limit = 1024
 
     @contextmanager
     def transaction(self):
@@ -214,7 +222,7 @@ class Database(object):
                     raise
                 logger.info('%s: after %s retries, retrying', e, tries)
                 time.sleep(tries * RETRY_BACKOFF_FACTOR)
-                self.reopen(refresh=refresh)
+                self.reopen(refresh_if_needed=refresh)
                 tries += 1
 
     def __init__(self, path,
@@ -228,6 +236,8 @@ class Database(object):
         self.spelling = spelling
         self.replicated = replicated
         self.type_map = {}
+        self.parsers_by_language = {}
+        self.query_cache = LRUDict(limit=self.query_cache_limit)
 
         if isinstance(path, basestring):
             if writable:
@@ -277,7 +287,7 @@ class Database(object):
 
     def check_prefix(self, name, boolean=False):
         prefixes = set(self.relevance_prefixes.keys())
-        prefixes.union(self.boolean_prefixes.keys())
+        prefixes = prefixes.union(self.boolean_prefixes.keys())
         if name not in prefixes:
             upped = _prefix(name)
             if boolean:
@@ -288,6 +298,8 @@ class Database(object):
     def add_prefix(self, key, value, ignore_duplicates=True):
         """Add a prefix mapping to the database.
         """
+        # FIXME:
+        # Why is it only checking boolean prefixes and then adding to relevance?
         if key in self.boolean_prefixes:
             if ignore_duplicates:
                 logger.warning('Conflicting relevance prefix %s', key)
@@ -490,10 +502,11 @@ class Database(object):
                 term, typ = term_tup
             if ':' in term:
                 prefix, _, value = term.partition(':')
+                # FIXME: check_prefix has a second param called boolean=False
+                # Is passing the typ in the right thing to do?
                 self.check_prefix(prefix, typ)
                 term = 'X%s:%s' % (prefix.upper(), value)
             doc.add_term(term, wdfinc)
-                
         for post in data.get('posts', ()):
             pass
 
@@ -520,7 +533,7 @@ class Database(object):
                     self.backend.add_spelling('food')
                     tg.set_flags(xapian.TermGenerator.FLAG_SPELLING)
                 except:
-                    pass # noop for backends that don't support spelling
+                    pass  # noop for backends that don't support spelling
             if lang in snowball.stoppers:
                 tg.set_stemmer(xapian.Stem(lang))
                 tg.set_stopper(snowball.stoppers[lang])
@@ -558,7 +571,7 @@ class Database(object):
             doc.set_data(data)
         return doc
 
-    def reopen(self, retry_limit=RETRY_LIMIT, refresh=True):
+    def reopen(self, retry_limit=RETRY_LIMIT, refresh_if_needed=True):
         """
         Reopen the database.  Called before most query methods.  If
         replication is used, the db is closed and reopened.
@@ -573,43 +586,8 @@ class Database(object):
             self.close()
             self.backend = xapian.Database(self.db_path)
 
-        if refresh:
-            try:
-                op = lambda: self.backend.metadata_keys()
-                # don't recurse into refresh here, just reopen and retry
-                keys = self.retry_if_modified(op, retry_limit, False)
-            except UnimplementedError:
-                # Inmemory backends don't expose metadata, so just return
-                # since they can't persist anything anyway!
-                return
-
-            self.relevance_prefixes = {}
-            self.boolean_prefixes = {}
-            self.values = {}
-            self.value_sorts = {}
-
-            for k in keys:
-                op = lambda: self.backend.get_metadata(k)
-                val = self.retry_if_modified(op, retry_limit, False)
-                if k.startswith(self.relevance_prefix):
-                    prefix = k[len(self.relevance_prefix):]
-                    self.relevance_prefixes[prefix] = val
-                elif k.startswith(self.boolean_prefix):
-                    prefix = k[len(self.boolean_prefix):]
-                    self.boolean_prefixes[prefix] = val
-                elif k.startswith(self.value_prefix):
-                    value = k[len(self.value_prefix):]
-                    self.values[value] = int(val)
-                elif k.startswith(self.value_sort_prefix):
-                    value = k[len(self.value_sort_prefix):]
-                    self.value_sorts[value] = val
-            try:
-                # hit the property to refresh this value
-                count = self.value_count
-                count = count # pyflakes
-            except ValueError:
-                if self._writable:
-                    self.value_count = 0
+        if refresh_if_needed and self.is_metadata_changed:
+            self.meta_refresh()
 
     def begin(self):
         if self._writable:
@@ -617,7 +595,7 @@ class Database(object):
             try:
                 self.backend.begin_transaction()
             except Exception:
-                pass # noop for backends that don't support transactions
+                pass  # noop for backends that don't support transactions
 
     def cancel(self):
         if self._writable:
@@ -683,6 +661,53 @@ class Database(object):
             qp.set_stemming_strategy(QueryParser.STEM_SOME)
         return qp
 
+    @property
+    def is_metadata_changed(self):
+        return self._metadata_keyset != self._get_metadata_keyset()
+
+    def _get_metadata_keyset(self, retry_limit=RETRY_LIMIT):
+        try:
+            op = lambda: set(self.backend.metadata_keys())
+            # don't recurse into refresh here, just reopen and retry
+            return self.retry_if_modified(op, retry_limit, False)
+        except UnimplementedError:
+            # Inmemory backends don't expose metadata, so just return
+            # since they can't persist anything anyway!
+            pass
+        return
+
+    def meta_refresh(self, retry_limit=RETRY_LIMIT):
+        self.parsers_by_language = {}
+        self.relevance_prefixes = {}
+        self.boolean_prefixes = {}
+        self.values = {}
+        self.value_sorts = {}
+        self.query_cache = {}
+
+        self._metadata_keyset = self._get_metadata_keyset()
+        for k in list(self._metadata_keyset):
+            op = lambda: self.backend.get_metadata(k)
+            val = self.retry_if_modified(op, retry_limit, False)
+            if k.startswith(self.relevance_prefix):
+                prefix = k[len(self.relevance_prefix):]
+                self.relevance_prefixes[prefix] = val
+            elif k.startswith(self.boolean_prefix):
+                prefix = k[len(self.boolean_prefix):]
+                self.boolean_prefixes[prefix] = val
+            elif k.startswith(self.value_prefix):
+                value = k[len(self.value_prefix):]
+                self.values[value] = int(val)
+            elif k.startswith(self.value_sort_prefix):
+                value = k[len(self.value_sort_prefix):]
+                self.value_sorts[value] = val
+        try:
+            # hit the property to refresh this value
+            count = self.value_count
+            count = count  # pyflakes
+        except ValueError:
+            if self._writable:
+                self.value_count = 0
+
     def querify(self, query,
                 language=None,
                 translit=None,
@@ -706,13 +731,24 @@ class Database(object):
             if query == "":
                 return Query("")
             else:
+                cache_key = (query, language, translit, default_op, parser_flags)
+                if cache_key in self.query_cache:
+                    return self.query_cache[cache_key]
                 query = query.lower()
                 if translit:
                     query = query.encode(translit)
-                def op():
-                    qp = self.prepare_query_parser(language, default_op)
+                qp = self.parsers_by_language.get(language)
+                if not qp:
+                    def prepare_op():
+                        return self.prepare_query_parser(language, default_op)
+                    qp = self.retry_if_modified(prepare_op, retry_limit)
+                    self.parsers_by_language[language] = qp
+
+                def query_op():
                     return qp.parse_query(query, parser_flags)
-                return self.retry_if_modified(op, retry_limit)
+                result = self.retry_if_modified(query_op, retry_limit)
+                self.query_cache[cache_key] = result
+                return result
         else:
             return reduce(partial(Query, default_op),
                           (self.querify(q,
@@ -856,6 +892,8 @@ class Database(object):
             counter = self.count
 
         results = {}
+        query = self.querify(query, language=language)
+
         suggestions = self.suggest(query,
                                    prefix=prefix,
                                    language=language,
@@ -864,15 +902,19 @@ class Database(object):
                                    klimit=klimit,
                                    kmlimit=kmlimit,
                                    echo=echo,
-                                   retry_limit=retry_limit)
+                                   retry_limit=retry_limit,
+                                   format_term=False)
         for facet in suggestions:
-            q = self.querify([query, facet],
-                             retry_limit=retry_limit,
-                             language=language)
+            q = Query(Query.OP_AND, [query, facet])
             if echo:
                 print str(q)
-            else:
-                results[facet] = counter(q, language=language)
+            if prefix and facet.startswith('X%s:' % prefix.upper()):
+                suffix = facet[len(prefix) + 2:]
+                if ' ' in suffix or '..' in suffix:
+                    facet = '%s:"%s"' % (prefix, suffix)
+                else:
+                    facet = '%s:%s' % (prefix, suffix)
+            results[facet] = counter(q, language=language)
         return results
 
     def expand(self, query, expand,
@@ -945,7 +987,7 @@ class Database(object):
         enq.set_query(query)
         op = lambda: enq.get_mset(0, 0, limit)
         mset = self.retry_if_modified(op, retry_limit)
-            
+
         return mset.get_matches_estimated()
 
     def term_freq(self, term):
@@ -964,19 +1006,24 @@ class Database(object):
         Describe the parsed query.
         """
         def op():
-            qp = self.prepare_query_parser(language, default_op)
+            qp = self.parsers_by_language.get(language)
+            if not qp:
+                qp = self.prepare_query_parser(language, default_op)
             return qp.parse_query(query, default_parser_flags) 
         return str(self.retry_if_modified(op, retry_limit))
 
     def spell(self, query,
               language=None,
+              default_op=Query.OP_AND,
               retry_limit=RETRY_LIMIT):
         """
         Suggest a query string with corrected spelling.
         """
         self.reopen()
         def op():
-            qp = self.prepare_query_parser(language)
+            qp = self.parsers_by_language.get(language)
+            if not qp:
+                qp = self.prepare_query_parser(language, default_op)
             qp.parse_query(query, QueryParser.FLAG_SPELLING_CORRECTION)
             return qp.get_corrected_query_string().decode('utf8')
         return self.retry_if_modified(op, retry_limit)
@@ -996,7 +1043,8 @@ class Database(object):
                 echo=False,
                 default_op=Query.OP_AND,
                 parser_flags=default_parser_flags,
-                retry_limit=RETRY_LIMIT):
+                retry_limit=RETRY_LIMIT,
+                format_term=True):
         """
         Suggest terms that would possibly yield more relevant results
         for the given query.
@@ -1044,7 +1092,7 @@ class Database(object):
 
         for item in eset.items:
             val = item[0].decode('utf8')
-            if prefix and val.startswith('X%s:' % prefix.upper()):
+            if format_term and prefix and val.startswith('X%s:' % prefix.upper()):
                 suffix = val[len(prefix) + 2:]
                 if ' ' in suffix or '..' in suffix:
                     val = '%s:"%s"' % (prefix, suffix)
